//-----------------------------------------------------------------------
// Canthology: software that makes it easy to create anthologies.
//
// Copyright (C) 2011 Ciaran McHale.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------

package org.canthology.canthology;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import java.util.ArrayList;
import java.util.HashSet;


class Util
{

	public static void assertion(boolean b)
	{
		if (!b) {
			throw new AssertionError("assertion failed");
		}
	}


	public static void assertion(boolean b, String msg)
	{
		if (!b) {
			throw new IllegalStateException(
				"assertion failure: " + msg);
		}
	}


	public static boolean arrayContains(
		String[]				array,
		String					str)
	{
		int						i;

		for (i = 0; i < array.length; i++) {
			if (array[i].equals(str)) {
				return true;
			}
		}
		return false;
	}


	public static String[] concatenate(
		String[]				array1,
		String[]				array2)
	{
		String[]				result;
		int						i;
		int						resultIndex;

		result = new String[array1.length + array2.length];
		resultIndex = 0;
		for (i = 0; i < array1.length; i++) {
			result[resultIndex++] = array1[i];
		}
		for (i = 0; i < array2.length; i++) {
			result[resultIndex++] = array2[i];
		}
		return result;
	}


	public static String[] concatenate(
		String[]				array,
		ArrayList				list)
	{
		String[]				result;
		int						i;
		int						listLen;
		int						resultIndex;

		listLen = list.size();
		result = new String[array.length + listLen];
		resultIndex = 0;
		for (i = 0; i < array.length; i++) {
			result[resultIndex++] = array[i];
		}
		for (i = 0; i < listLen; i++) {
			result[resultIndex++] = (String)list.get(i);
		}
		return result;
	}


	public static String[] removeDuplicates(String[] strings)
	{
		ArrayList				result;
		HashSet					set;
		int						i;

		set = new HashSet();
		result = new ArrayList();
		for (i = 0; i < strings.length; i++) {
			if (!set.contains(strings[i])) {
				result.add(strings[i]);
			}
			set.add(strings[i]);
		}
		return (String[])result.toArray(new String[0]);
	}


	public static String readTextFile(String fileName)
	{
		BufferedReader			in;
		StringBuffer			result;
		String					line;

		result = new StringBuffer();
		try {
			in = new BufferedReader(new FileReader(fileName));
			while ((line = in.readLine()) != null) {
				result.append(line).append("\n");
			}
			in.close();
		} catch(IOException ex) {
			throw new CanthologyException("error reading file '" + fileName
										  + "': " + ex.getMessage());
		}
		return result.toString();
	}


	public static void writeTextFile(String fileName, String contents)
	{
		PrintWriter				out;
		String					line;

		try {
			out = new PrintWriter(new BufferedWriter(new FileWriter(fileName)));
			out.print(contents);
			out.close();
		} catch(IOException ex) {
			throw new CanthologyException("error writing file '" + fileName
										  + "': " + ex.getMessage());
		}
	}


	public static void copyBinaryFile(String fromFileName,String toFileName)
	{
		DataInputStream			in;
		DataOutputStream		out;
		byte[]					data;
		int						size;

		data = new byte[8 * 1024];
		try {
			out = new DataOutputStream(new BufferedOutputStream(
					new FileOutputStream(toFileName)));
			in = new DataInputStream(new BufferedInputStream(
					new FileInputStream(fromFileName)));
			while ((size = in.read(data)) != -1) {
				out.write(data, 0, size);
			}
			in.close();
			out.close();
		} catch(IOException ex) {
			throw new CanthologyException("error copying file '" + fromFileName
								+ "' to '" + toFileName + "': " + ex);
		}
	}


	public static String replace(
		String					origStr,
		String					searchStr,
		String					replacementStr)
	{
		StringBuffer			result;
		int						origStrLen;
		int						searchStrLen;
		int						currStart;
		int						pIndex;

		result = new StringBuffer();
		origStrLen = origStr.length();
		searchStrLen = searchStr.length();
		currStart = 0;
		pIndex = origStr.indexOf(searchStr, currStart);
		while (pIndex != -1) {
			result.append(origStr.substring(currStart, pIndex));
			result.append(replacementStr);
			currStart = pIndex + searchStrLen;
			pIndex = origStr.indexOf(searchStr, currStart);
		}
		result.append(origStr.substring(currStart));
		return result.toString();
	}


	public static String replace(
		String					origStr,
		String[]				searchAndReplacePairs)
	{
		int						i;
		int						currIndex;
		String					subStr;
		String					replaceStr;
		StringBuffer			result;
		SearchAndReplacePair[]	pairs;
		SearchAndReplacePair	nextPair;

		pairs = new SearchAndReplacePair[searchAndReplacePairs.length / 2];
		for (i = 0; i < searchAndReplacePairs.length; i += 2) {
			pairs[i/2] = new SearchAndReplacePair(origStr,
												  searchAndReplacePairs[i + 0],
												  searchAndReplacePairs[i + 1]);
		}

		result = new StringBuffer();
		currIndex = 0;
		nextPair = findNextPair(origStr, currIndex, pairs);
		while (nextPair != null) {
			//--------
			// Debug diagnostics
			//--------
			//System.out.println("replace()"
			//	+ ": currIndex=" + currIndex
			//	+ "; indexOf=" + nextPair.indexOf
			//	+ "; found='" + nextPair.search + "'");
			subStr = origStr.substring(currIndex,
						nextPair.indexOf);
			result.append(subStr);
			result.append(nextPair.replace);
			currIndex = nextPair.indexOf + nextPair.length;
			for (i = 0; i < pairs.length; i++) {
				pairs[i].findNext(currIndex);
			}
			nextPair = findNextPair(origStr, currIndex, pairs);
		}
		subStr = origStr.substring(currIndex);
		result.append(subStr);

		return result.toString();
	}


	private static SearchAndReplacePair findNextPair(
		String						origStr,
		int							currIndex,
		SearchAndReplacePair[]		pairs)
	{
		int							i;
		SearchAndReplacePair		bestSoFar;
		SearchAndReplacePair		item;

		bestSoFar = null;
		for (i = 0; i < pairs.length; i++) {
			item = pairs[i];
			if (item.indexOf == -1) {
				continue;
			}
			if (bestSoFar == null) {
				bestSoFar = item;
				continue;
			}
			if (bestSoFar.indexOf < item.indexOf) {
				continue;
			}
			if (bestSoFar.indexOf > item.indexOf) {
				bestSoFar = item;
				continue;
			}
			if (bestSoFar.length < item.length) {
				bestSoFar = item;
			}
		}
		return bestSoFar;
	}


	public static boolean execCommand(
		String						cmd,
		File						cwd,
		String						input,
		StringBuffer				output)
	{
		String						line;
		Process						p;
		BufferedReader				processStdout;
		StreamReaderThread			outThread;
		StreamReaderThread			errThread;
		StreamWriterThread			inThread;
		int							exitVal;

		//--------
		// Ensure that "output" is empty.
		//--------
		output.delete(0, output.length());

		//--------
		// Execute the command and capture its stdout
		// (and stderr too if the proces has a non-zero exit status).
		//
		// Note: the article "When Runtime.exec() won't" by Michael
		// C. Daconta, JavaWorld.com, 12/29/00, explains that the
		// stdout and stderr of an executed process should be
		// read concurrently by separate threads to avoid the
		// possibility of deadlock. You can find the article at:
		// www.javaworld.com/javaworld/jw-12-2000/jw-1229-traps.html
		//--------
		try {
			p = Runtime.getRuntime().exec(cmd, null, cwd);
		} catch(Exception ex) {
			output.append("cannot execute '" + cmd + "': " + ex.getMessage());
			return false;
		}
		outThread = new StreamReaderThread(p.getInputStream());
		errThread = new StreamReaderThread(p.getErrorStream());
		inThread  = new StreamWriterThread(input, p.getOutputStream());
		outThread.start();
		errThread.start();
		inThread.start();

		//--------
		// Wait the the child process and the in/out/err threads to die.
		//--------
		while (true) {
			try {
				exitVal = p.waitFor(); break;
			} catch(InterruptedException ex) { }
		}
		while (true) {
			try {
				inThread.join(); break;
			} catch(InterruptedException ex) { }
		}
		while (true) {
			try {
				outThread.join(); break;
			} catch(InterruptedException ex) { }
		}
		while (true) {
			try {
				errThread.join(); break;
			} catch(InterruptedException ex) { }
		}

		//--------
		// If successful, then return the child process's standard output.
		// Otherwise, return its standard input *and* error.
		//--------
		if (exitVal == 0) {
			output.append(outThread.getBuf());
		} else {
			output.append(outThread.getBuf());
			output.append("\n");
			output.append(errThread.getBuf());
		}
		return (exitVal == 0);
	}

}


class SearchAndReplacePair
{
	String		source;
	String		search;
	String		replace;
	int		length;
	int		indexOf;
	int		sourceLength;

	public SearchAndReplacePair(String source, String search, String replace)
	{
		this.source = source;
		this.sourceLength = source.length();
		this.search  = search;
		this.replace = replace;
		this.length  = search.length();
		this.indexOf = source.indexOf(search);
	}

	public void findNext(int fromIndex)
	{
		if (indexOf == -1 || indexOf + 1 == sourceLength) {
			indexOf = -1;
		} else {
			indexOf = source.indexOf(search, fromIndex);
		}
	}

}


class StreamReaderThread extends Thread
{
	StreamReaderThread(InputStream inputStream)
	{
		this.inputStream = inputStream;
		this.buf = new StringBuffer();
	}

	String	getBuf()
	{
		return buf.toString();
	}

	public void run()
	{
		InputStreamReader	isr;
		BufferedReader		br;
		boolean				first;
		String				line;

		try {
			isr = new InputStreamReader(inputStream);
			br = new BufferedReader(isr);
			first = true;
			while ((line = br.readLine()) != null) {
				if (!first) {
					buf.append("\n");
				}
				buf.append(line);
				first = false;
			}
		} catch(IOException ex) {
			ex.printStackTrace();
		}
	}

	private InputStream		inputStream;
	private StringBuffer	buf;
}


class StreamWriterThread extends Thread
{
	StreamWriterThread(String text, OutputStream outputStream)
	{
		this.outputStream = outputStream;
		this.text = text;
	}

	public void run()
	{
		OutputStreamWriter	out;

		//--------
		// Send text to the standard input of the process.
		//--------
		try {
			out = new OutputStreamWriter(outputStream);
			out.write(text);
			out.close();
		} catch(IOException ex) {
			//ex.printStackTrace();
		}
	}

	private OutputStream	outputStream;
	private String		text;
}

